"""Plugin functionality for adding function estimators.

This package contains the abstract base classes for function estimator plugins,
and the default function estimators that are part of `ropt`.
"""
